\
\ Template code for CGIshell, RTK, 09-Oct-95
\
\ last mod: 24-Oct-95
\

( yet more disk I/O words by C. Heilman )

\ create space for the fcb and a word to access it
variable FCB 78 allot  ( our File's Control Block )
: +FCB ( offset -- addr ) fcb + ;  ( offset into fcb )

\ setup for a (register based) file manager toolbox call
: FTRAP ( -- ) fcb >abs  ,$ 205E ;  ( movea.l [ps]+,a0 )

: CLOSE ( -- ) ftrap ,$ A001  ftrap ,$ A013 ;  ( _Close & _FlushBuffer )
: ?DERROR ( -- ) \ report error if result is not zero
    16 +fcb @ ?dup IF  ." DiskError" .  close  abort THEN ;

\ open a file with the address of a string of the pathname on the stack
: OPEN ( addr -- ) \ addr is a Forth style string - str[255]
    fcb 80 0 fill           \   clear the fcb for a new file
    >abs  18 +fcb  2!        \  set name of the file to string
    ftrap ,$ A000  ?derror ;  \ _Open the file in the fcb

\ create a file
: NEWFILE ( name.addr -- )
    fcb 80 0 fill           \   clear the fcb for a new file
    >abs  18 +fcb  2!        \  set name of the file to string
    ftrap ,$ A008  ?derror    ( _Create )
    ,s TEXT 32 +fcb 2!         \ TEXT type
    ftrap ,$ A00D ?derror ;   ( _SetFileInfo )

\ return the filesize !!! MUST BE <32K !!!
: @SIZE ( -- bytes ) ftrap ,$ A011  30 +fcb @ ;  ( _GetEOF )

\ set some fcb parameters
: !SIZE ( bytes -- ) 38 +fcb ! ;      \ set bytes-to-read/write
: !BUFF ( addr -- ) >abs 32 +fcb 2! ;  \ set read/write buffer pointer

\ read/write with buffer addr and bytes to read/write on the stack
: READ ( addr count -- )  !size !buff ftrap ,$ A002  ?derror ;  ( _Read )
: WRITE ( addr count -- ) !size !buff ftrap ,$ A003  ?derror ;  ( _Write )

\ read/write file a byte at a time to/from the stack
: GETCHR ( -- c ) here 1 read  here c@ ;
: PUTCHR ( c -- ) here c!  here 1 write ;

\ read until character (c) is encountered
: CREAD ( addr c -- bytes_read )
    44 +fcb c!  128 45 +fcb c!  \ setup ioPosMode
    @SIZE read  42 +fcb @ ;     \ put lowbyte of ioActCount on stack

\ A defining word for strings
: $[  \ compiling: ( -- ) enclose a ] terminated string
    CREATE  93 word here  c@ 1+ dup 2 mod +  allot
    DOES>  ;  \ runtime action: ( -- addr ) \ <<-- no count!

( end I/O words )

1024 constant buffSize        \ size of input buffer
create rwbuff buffSize allot  \ the buffer
variable ~i                   \ start by filling the buffer
variable ~f                   \ number of bytes into the file

: fillbuff ( -- )  \ call this after OPEN to use the buffer
   1024 ~i !  0 ~f !  rwbuff buffSize 0 fill ;

: getch ( -- c )  \ read a character from the buffer
   ~i @ dup 1024 < IF
     rwbuff + c@  1 ~i +!  1 ~f +!
   ELSE
     drop 0 ~i !  \ clear index
     rwbuff buffSize 0 fill
     @size  ~f @ - 3 -  dup 1 < IF
       drop  -1  \ nothing to read, return -1
     ELSE
       buffSize  min  \ # bytes to read
       rwbuff swap READ
       rwbuff c@  1 ~i +!
     THEN
   THEN ;

( string array words )

: array>> ( #elements --  )  \ create an array of #elements
   create 2* allot ;

: !array ( data index array -- )  \ store in index
   swap 2* + ! ;

: @array ( index array -- data )  \ get data in an array
   swap 2* + @ ;

: >array ( 00 01 .. n array -- )  \ store entire array
   swap -1 swap 1- DO
     dup rot swap r swap !array
   -1 +LOOP ;

\ add these to the number convert

: n>str ( n s -- )  \ integer to string
   >r 0 d>f r> 0 fix f>str ;

: str>num ( s -- n )  \ string to integer
   str>f f>d drop ;


variable ~output  \ output string
variable ~array   \ array of substitute strings
variable ~fname   \ filename string

: >digit ( c -- n ) 48 - ; \ character to number

variable ~length   \ hold length of output string
: >>output  ( c -- )  \ append a character
  ~output @ ~length @ + dup >r c!  0 r> 1+ c! 1 ~length +! ; macro

: template ( output array fname new|append -- )
   \ process a template file
   >r  ~fname !  ~array !  ~output !  r>
   IF 0 ~output @ c!  0 ~length !        \ zero the string
   ELSE ~output @ length ~length ! THEN  \ append to string 
   ~fname @ open                         \ open the file
   fillbuff                              \ init input buffer
   @size 0 DO   
     getch                               \ get a character
     dup 96 = IF                         \ check for ` (backquote)
      drop                               \ lose the `
      getch >digit 10 *                  \ do the substitution
      getch >digit +                     \ calculate index
      ~array @  @array                   \ get address of string
      ~output @ strcpy                   \ and put it in
      ~output @ length ~length !         \ adjust length
      3
     ELSE
      dup 31 > IF
       >>output                          \ append to output
      ELSE drop THEN 1
     THEN
   +LOOP
   close                                 \ close the file
;
